#include <math.h>
#include <string.h>
#include <tk.h>
#include "cairotcl.h"
#include <cairo.h>
#include <cairo-pdf.h>
#include <cairo-svg.h>

struct SurfaceState {
  Tcl_HashTable hash;
  int uid;
} surfaceState;

struct SurfaceData {
  cairo_surface_t *surface;
  cairo_t *context;
};

static void UpdateStringOfPath(Tcl_Obj *obj);
static Tcl_ObjType PathType = {
  "path",
  NULL,
  NULL,
  UpdateStringOfPath,
  NULL
};

static void UpdateStringOfPath(Tcl_Obj *obj) {
  static int pathCount = 0;
  char buffer[100];
  int len;

  len = sprintf(buffer, "path%d", pathCount);
  //fprintf(stderr, "updateString %p %s\n", obj, buffer);
  pathCount++;
  obj->bytes = ckalloc(len+1);
  strcpy(obj->bytes, buffer);
  obj->length = len;
}

Tcl_Obj *newPathObj(cairo_path_t *path) {
  Tcl_Obj *obj;

  obj = Tcl_NewObj();
  obj->bytes = NULL;
  obj->internalRep.otherValuePtr = path;
  obj->typePtr = &PathType;
  //fprintf(stderr, "newPathObj %p %p len: %d\n", obj, path, path->num_data);
  return obj;
}

cairo_path_t *getPathFromObj(Tcl_Obj *obj) {
  cairo_path_t *path;
  if (obj->typePtr != &PathType)
    return NULL;
  path = (cairo_path_t *)obj->internalRep.otherValuePtr;
  //fprintf(stderr, "getPath %p %p len: %d\n", obj, path, path->num_data);
  return path;
}

int newPath(Tcl_Interp *interp, cairo_t *cr)
{
  cairo_new_path(cr);
  return TCL_OK;
}

int moveTo(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x = 0, y = 0;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 2, objv, "x y");
    return TCL_ERROR;
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x);
  Tcl_GetDoubleFromObj(interp, objv[3], &y);
  cairo_move_to(cr, x, y);
  return TCL_OK;
}

int lineTo(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x = 0, y = 0;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 2, objv, "x y");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x);
  Tcl_GetDoubleFromObj(interp, objv[3], &y);
  cairo_line_to(cr, x, y);
  return TCL_OK;
}

int line(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  Tcl_Obj *obj;
  int i;
  int len = 0;
  double x = 0, y = 0;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "coords");
  }
  Tcl_ListObjLength(interp, objv[2], &len);
  if (len < 4 || len % 2 != 0) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong coord length", -1));
    return TCL_ERROR;
  }
  for (i = 0; i < len; i += 2) {
    Tcl_ListObjIndex(interp, objv[2], i, &obj);
    Tcl_GetDoubleFromObj(interp, obj, &x);
    Tcl_ListObjIndex(interp, objv[2], i+1, &obj);
    Tcl_GetDoubleFromObj(interp, obj, &y);
    if (i == 0) {
      cairo_move_to(cr, x, y);
    } else {
      cairo_line_to(cr, x, y);
    }
  }
  return TCL_OK;
}

int rectangle(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x0 = 0, y0 = 0, x1 = 0, y1 = 0;
  if (objc != 6) {
    Tcl_WrongNumArgs(interp, 2, objv, "x0 y0 x1 y1");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x0);
  Tcl_GetDoubleFromObj(interp, objv[3], &y0);
  Tcl_GetDoubleFromObj(interp, objv[4], &x1);
  Tcl_GetDoubleFromObj(interp, objv[5], &y1);
  cairo_rectangle(cr, x0, y0, x1, y1);
  return TCL_OK;
}

int arc(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x = 0, y = 0, r = 0, phi0 = 0, phi1 = 0;
  if (objc != 7) {
    Tcl_WrongNumArgs(interp, 2, objv, "x y r phi0 phi1");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x);
  Tcl_GetDoubleFromObj(interp, objv[3], &y);
  Tcl_GetDoubleFromObj(interp, objv[4], &r);
  Tcl_GetDoubleFromObj(interp, objv[5], &phi0);
  Tcl_GetDoubleFromObj(interp, objv[6], &phi1);
  cairo_arc(cr, x, y, r, phi0, phi1);
  return TCL_OK;
}

int circle(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x = 0, y = 0, r = 0;
  if (objc != 5) {
    Tcl_WrongNumArgs(interp, 2, objv, "x y r");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x);
  Tcl_GetDoubleFromObj(interp, objv[3], &y);
  Tcl_GetDoubleFromObj(interp, objv[4], &r);
  cairo_arc(cr, x, y, r, 0.0, 2*M_PI);
  return TCL_OK;
}

int curveTo(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double x0 = 0, y0 = 0;
  double x1 = 0, y1 = 0;
  double x2 = 0, y2 = 0;
  if (objc != 8) {
    Tcl_WrongNumArgs(interp, 2, objv, "x0 y0 x1 y1 x2 y2");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &x0);
  Tcl_GetDoubleFromObj(interp, objv[3], &y0);
  Tcl_GetDoubleFromObj(interp, objv[4], &x1);
  Tcl_GetDoubleFromObj(interp, objv[5], &y1);
  Tcl_GetDoubleFromObj(interp, objv[6], &x2);
  Tcl_GetDoubleFromObj(interp, objv[7], &y2);
  cairo_curve_to(cr, x0, y0, x1, y1, x2, y2);
  return TCL_OK;
}

int curve(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  double *x, *y;
  int i;
  int len;
  Tcl_Obj *obj;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "coords");
  }
  Tcl_ListObjLength(interp, objv[2], &len);
  if (len < 8 || (len-2) % 6 != 0) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("wrong coord length", -1));
    return TCL_ERROR;
  }
  len /= 2;
  x = (double *)ckalloc(len*sizeof(double));
  y = (double *)ckalloc(len*sizeof(double));
  for (i = 0; i < len; i++) {
    Tcl_ListObjIndex(interp, objv[2], 2*i, &obj);
    Tcl_GetDoubleFromObj(interp, obj, &x[i]);
    Tcl_ListObjIndex(interp, objv[2], 2*i+1, &obj);
    Tcl_GetDoubleFromObj(interp, obj, &y[i]);
    //fprintf(stderr, "%.1f %.1f\n", x[i], y[i]);
  }
  cairo_move_to(cr, x[0], y[0]);
  for (i = 1; i < len; i += 3) {
    //fprintf(stderr, "i: %d\n", i);
    cairo_curve_to(cr, x[i], y[i], x[i+1], y[i+1], x[i+2], y[i+2]);
  }
  return TCL_OK;
}

int setSourceRgb(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double r = 0, g = 0, b = 0;
  if (objc != 5) {
    Tcl_WrongNumArgs(interp, 2, objv, "red green blue");
    return TCL_ERROR;
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &r);
  Tcl_GetDoubleFromObj(interp, objv[3], &g);
  Tcl_GetDoubleFromObj(interp, objv[4], &b);
  cairo_set_source_rgb(cr, r, g, b);
  return TCL_OK;
}

int setSourceRgba(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double r = 0, g = 0, b = 0, a = 0;
  if (objc != 6) {
    Tcl_WrongNumArgs(interp, 2, objv, "red green blue alpha");
    return TCL_ERROR;
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &r);
  Tcl_GetDoubleFromObj(interp, objv[3], &g);
  Tcl_GetDoubleFromObj(interp, objv[4], &b);
  Tcl_GetDoubleFromObj(interp, objv[5], &a);
  cairo_set_source_rgba(cr, r, g, b, a);
  return TCL_OK;
}

int setLineCap(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  static const char *capTypes[] = {
    "butt", "round", "square", NULL
  };
  int index;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "capType");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[2], capTypes, "capType", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  cairo_set_line_cap(cr, index);
  return TCL_OK;
}

int setLineJoin(Tcl_Interp *interp, cairo_t *cr, int objc, Tcl_Obj *CONST objv[])
{
  static const char *capTypes[] = {
    "miter", "round", "bevel", NULL
  };
  int index;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "joinType");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[2], capTypes, "capType", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  cairo_set_line_join(cr, index);
  return TCL_OK;
}

int setLineWidth(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double w = 0;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "width");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &w);
  cairo_set_line_width(cr, w);
  return TCL_OK;
}

int setDash(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double *dashes = 0;
  int i, len;
  double offset = 0;
  Tcl_Obj *obj;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "dashlist ?offset?");
  }
  Tcl_ListObjLength(interp, objv[2], &len);
  dashes = (double *)Tcl_Alloc(len*sizeof(double));
  for (i = 0; i < len; i++) {
    Tcl_ListObjIndex(interp, objv[2], i, &obj);
    Tcl_GetDoubleFromObj(interp, obj, &dashes[i]);
  }
  if (objc > 3)
    Tcl_GetDoubleFromObj(interp, objv[3], &offset);
  cairo_set_dash(cr, dashes, len, offset);
  Tcl_Free((char *)dashes);
  return TCL_OK;
}

int selectFontFace(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  static const char *fontSlants[] = {"normal", "italic", "oblique"};
  static const char *fontWeights[] = {"normal", "bold"};
  int idxSlant, idxWeight;

  if (objc != 5) {
    Tcl_WrongNumArgs(interp, 2, objv, "family slant weight");
  }
  if (Tcl_GetIndexFromObj(interp, objv[3], fontSlants, "slant", 0, &idxSlant)
      != TCL_OK) {
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[4], fontWeights, "weight", 0, &idxWeight)
      != TCL_OK) {
    return TCL_ERROR;
  }
  cairo_select_font_face(cr, Tcl_GetString(objv[2]), idxSlant, idxWeight);
  return TCL_OK;
}

int setFontSize(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double size = 0;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "size");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &size);
  cairo_set_font_size(cr, size);
  return TCL_OK;
}

int showText(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "text");
  }
  cairo_show_text(cr, Tcl_GetString(objv[2]));
  return TCL_OK;
}

int textPath(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "text");
  }
  cairo_text_path(cr, Tcl_GetString(objv[2]));
  return TCL_OK;
}

int textExtents(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  cairo_text_extents_t extents;
  Tcl_Obj *res;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "text");
  }
  cairo_text_extents(cr, Tcl_GetString(objv[2]), &extents);
  res = Tcl_NewListObj(0, NULL);
  Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(extents.width));
  Tcl_ListObjAppendElement(interp, res, Tcl_NewDoubleObj(extents.height));
  Tcl_SetObjResult(interp, res);
  return TCL_OK;
}

int scale(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double scx = 1;
  double scy = 1;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 2, objv, "scalex scaley");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &scx);
  Tcl_GetDoubleFromObj(interp, objv[3], &scy);
  cairo_scale(cr, scx, scy);
  return TCL_OK;
}

int rotate(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double phi = 0;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "angle");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &phi);
  cairo_rotate(cr, phi);
  return TCL_OK;
}

int translate(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double dx = 0;
  double dy = 0;
  if (objc != 4) {
    Tcl_WrongNumArgs(interp, 2, objv, "dx dy");
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &dx);
  Tcl_GetDoubleFromObj(interp, objv[3], &dy);
  cairo_translate(cr, dx, dy);
  return TCL_OK;
}

int copyPath(Tcl_Interp *interp, cairo_t *cr) {
  cairo_path_t *pathPtr;

  pathPtr = cairo_copy_path(cr);
  Tcl_SetObjResult(interp, newPathObj(pathPtr));
  return TCL_OK;
}

int appendPath(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  cairo_path_t *path;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "path");
  }
  path = getPathFromObj(objv[2]);
  if (path != NULL)
    cairo_append_path(cr, path);
  return TCL_OK;
}

int destroyPath(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  cairo_path_t *path;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "path");
  }
  path = getPathFromObj(objv[2]);
  cairo_path_destroy(path);
  objv[2]->internalRep.otherValuePtr = NULL;
  return TCL_OK;
}

int setOperator(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  static const char *operators[] = {
    "clear", "source", "over", "in", "out", "atop",
    "dest", "dest_over", "dest_in", "dest_out", "dest_atop",
    "xor", "add", "saturate", NULL
  };
  int cairoOperator;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "operator");
  }
  if (Tcl_GetIndexFromObj(interp, objv[2], operators, "operator", 0,
      &cairoOperator) != TCL_OK) {
    return TCL_ERROR;
  }
  cairo_set_operator(cr, cairoOperator);
  return TCL_OK;
}

int setFillRule(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  static const char *rules[] = {
    "winding", "even_odd", NULL
  };
  int fillRule;

  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "fillRule");
  }
  if (Tcl_GetIndexFromObj(interp, objv[2], rules, "fill_rule", 0,
      &fillRule) != TCL_OK) {
    return TCL_ERROR;
  }
  cairo_set_fill_rule(cr, fillRule);
  return TCL_OK;
}

/**
 * cr image image x y
 */
int drawImage(Tcl_Interp *interp, cairo_t *cr, int objc,
    Tcl_Obj *CONST objv[])
{
  double x, y;
  Tk_PhotoHandle tkImg;
  Tk_PhotoImageBlock block;
  unsigned char *pixelPtr;
  int w, h, off;
  int stride;
  cairo_surface_t *img;

  tkImg = Tk_FindPhoto(interp, Tcl_GetString(objv[2]));
  Tcl_GetDoubleFromObj(interp, objv[3], &x);
  Tcl_GetDoubleFromObj(interp, objv[4], &y);
  if (tkImg == 0)
    return TCL_ERROR;
  Tk_PhotoGetImage(tkImg, &block);
  stride = cairo_format_stride_for_width(CAIRO_FORMAT_ARGB32, block.width);
  pixelPtr = (unsigned char *)Tcl_Alloc(stride*block.height);
  for (h = 0; h < block.height; h++) {
    for (w = 0; w < block.width; w++) {
      for (off = 0; off < 3; off++) {
	pixelPtr[h*stride+w*4+off]
	  = block.pixelPtr[h*block.pitch+w*block.pixelSize+block.offset[2-off]];
      }
      pixelPtr[h*stride+w*4+3]
	  = block.pixelPtr[h*block.pitch+w*block.pixelSize+block.offset[3]];
    }
  }
  img = cairo_image_surface_create_for_data(pixelPtr, CAIRO_FORMAT_ARGB32,
    block.width, block.height, stride);
  cairo_set_source_surface(cr, img, x-block.width/2, y-block.height/2);
  cairo_paint(cr);
  cairo_surface_destroy(img);
  Tcl_Free((char *)pixelPtr);
  return TCL_OK;
}

int status(Tcl_Interp *interp, cairo_t *cr) {
  cairo_status_t status;
  status = cairo_status(cr);
  if (status == CAIRO_STATUS_SUCCESS) {
    return TCL_OK;
  }
  Tcl_SetObjResult(interp, Tcl_NewStringObj(cairo_status_to_string(status),
    -1));
  return TCL_ERROR;
}

int contextObjCmd(ClientData data, Tcl_Interp *interp, int objc,
  Tcl_Obj *CONST objv[])
{
  struct SurfaceData *surfacePtr = (struct SurfaceData *)data;
  static const char *contextCmds[] = {
    "new_path",
    "move_to", "line_to", "line", "rectangle", "arc", "circle",
    "curve_to", "curve", "close_path",
    "set_source_rgb", "set_source_rgba",
    "set_line_cap", "set_line_join", "set_line_width", "set_dash",
    "stroke", "stroke_preserve", "fill", "fill_preserve",
    "image",
    "set_fill_rule",
    "paint", "clip", "clip_preserve",
    "select_font_face", "set_font_size", "show_text", "text_path",
    "text_extents",
    "save", "restore", "show_page",
    "scale", "rotate", "translate",
    "copy_path", "append_path", "path_destroy",
    "push_group", "pop_group_to_source",
    "set_operator",
    "status",
    NULL
  };
  enum ctxCmdIdx {
    NEW_PATH_IX,
    MOVE_TO_IX, LINE_TO_IX, LINE_IX, RECTANGLE_IX, ARC_IX, CIRCLE_IX,
    CURVE_TO_IX, CURVE_IX, CLOSE_PATH_IX,
    SET_SOURCE_RGB_IX, SET_SOURCE_RGBA_IX,
    SET_LINE_CAP_IX, SET_LINE_JOIN_IX, SET_LINE_WIDTH_IX, SET_DASH_IX,
    STROKE_IX, STROKE_PRESERVE_IX, FILL_IX, FILL_PRESERVE_IX,
    IMAGE_IX,
    SET_FILL_RULE_IX,
    PAINT_IX, CLIP_IX, CLIP_PRESERVE_IX,
    SELECT_FONT_FACE_IX, SET_FONT_SIZE_IX, SHOW_TEXT_IX, TEXT_PATH_IX,
    TEXT_EXTENTS_IX,
    SAVE_IX, RESTORE_IX, SHOW_PAGE_IX,
    SCALE_IX, ROTATE_IX, TRANSLATE_IX,
    COPY_PATH_IX, APPEND_PATH_IX, DESTROY_PATH_IX,
    PUSH_GROUP_IX, POP_GROUP_TO_SOURCE_IX,
    SET_OPERATOR_IX,
    STATUS_IX
  };
  int index;
  cairo_t *cr;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], contextCmds, "command", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  cr = surfacePtr->context;
  switch (index) {
    case NEW_PATH_IX:
      return newPath(interp, cr);
    case MOVE_TO_IX:
      return moveTo(interp, cr, objc, objv);
    case LINE_TO_IX:
      return lineTo(interp, cr, objc, objv);
    case LINE_IX:
      return line(interp, cr, objc, objv);
    case RECTANGLE_IX:
      return rectangle(interp, cr, objc, objv);
    case ARC_IX:
      return arc(interp, cr, objc, objv);
    case CIRCLE_IX:
      return circle(interp, cr, objc, objv);
    case CURVE_TO_IX:
      return curveTo(interp, cr, objc, objv);
    case CLOSE_PATH_IX:
      cairo_close_path(cr);
      return TCL_OK;
    case CURVE_IX:
      return curve(interp, cr, objc, objv);
    case SET_SOURCE_RGB_IX:
      return setSourceRgb(interp, cr, objc, objv);
    case SET_SOURCE_RGBA_IX:
      return setSourceRgba(interp, cr, objc, objv);
    case SET_LINE_CAP_IX:
      return setLineCap(interp, cr, objc, objv);
    case SET_LINE_JOIN_IX:
      return setLineJoin(interp, cr, objc, objv);
    case SET_LINE_WIDTH_IX:
      return setLineWidth(interp, cr, objc, objv);
    case SET_DASH_IX:
      return setDash(interp, cr, objc, objv);
    case STROKE_IX:
      cairo_stroke(cr);
      return TCL_OK;
    case STROKE_PRESERVE_IX:
      cairo_stroke_preserve(cr);
      return TCL_OK;
    case FILL_IX:
      cairo_fill(cr);
      return TCL_OK;
    case FILL_PRESERVE_IX:
      cairo_fill_preserve(cr);
      return TCL_OK;
    case IMAGE_IX:
      drawImage(interp, cr, objc, objv);
      return TCL_OK;
    case SET_FILL_RULE_IX:
      return setFillRule(interp, cr, objc, objv);
    case PAINT_IX:
      cairo_paint(cr);
      return TCL_OK;
    case CLIP_IX:
      cairo_clip(cr);
      return TCL_OK;
    case CLIP_PRESERVE_IX:
      cairo_clip_preserve(cr);
      return TCL_OK;
    case SELECT_FONT_FACE_IX:
      return selectFontFace(interp, cr, objc, objv);
    case SET_FONT_SIZE_IX:
      return setFontSize(interp, cr, objc, objv);
    case SHOW_TEXT_IX:
      return showText(interp, cr, objc, objv);
    case TEXT_PATH_IX:
      return textPath(interp, cr, objc, objv);
    case TEXT_EXTENTS_IX:
      return textExtents(interp, cr, objc, objv);
    case SAVE_IX:
      cairo_save(surfacePtr->context);
      return TCL_OK;
    case RESTORE_IX:
      cairo_restore(surfacePtr->context);
      return TCL_OK;
    case SHOW_PAGE_IX:
      cairo_show_page(surfacePtr->context);
      return TCL_OK;
    case SCALE_IX:
      return scale(interp, cr, objc, objv);
    case ROTATE_IX:
      return rotate(interp, cr, objc, objv);
    case TRANSLATE_IX:
      return translate(interp, cr, objc, objv);
    case COPY_PATH_IX:
      return copyPath(interp, cr);
    case APPEND_PATH_IX:
      return appendPath(interp, cr, objc, objv);
    case DESTROY_PATH_IX:
      return destroyPath(interp, cr, objc, objv);
    case PUSH_GROUP_IX:
      cairo_push_group(cr);
      return TCL_OK;
    case POP_GROUP_TO_SOURCE_IX:
      cairo_pop_group_to_source(cr);
      return TCL_OK;
    case SET_OPERATOR_IX:
      return setOperator(interp, cr, objc, objv);
    case STATUS_IX:
      return status(interp, cr);
  }
  return TCL_OK;
}

int toPng(Tcl_Interp *interp, struct SurfaceData *surfacePtr,
  int objc, Tcl_Obj *CONST objv[]) {

  char *filename;
  if (objc != 3) {
    Tcl_WrongNumArgs(interp, 2, objv, "filename");
    return TCL_ERROR;
  }
  filename = Tcl_GetString(objv[2]);
  cairo_surface_write_to_png(surfacePtr->surface, filename);
  return TCL_OK;
}

int pdfSetSize(Tcl_Interp *interp, struct SurfaceData *surfacePtr,
  int objc, Tcl_Obj *CONST objv[]) {

  double scale;
  double width;
  double height;
  char *unit;

  unit = "pt";
  scale = 1.0;
  if (objc < 4) {
    Tcl_WrongNumArgs(interp, 2, objv, "width height ?pt|mm|cm?");
    return TCL_ERROR;
  }
  Tcl_GetDoubleFromObj(interp, objv[2], &width);
  Tcl_GetDoubleFromObj(interp, objv[3], &height);
  if (objc > 4)
    unit = Tcl_GetString(objv[4]);
  if (strcmp(unit, "mm") == 0) {
    scale = 72.0/25.4;
  }
  else if (strcmp(unit, "cm") == 0) {
    scale = 72.0/2.54;
  }
  cairo_pdf_surface_set_size(surfacePtr->surface, scale*width, scale*height);
  return TCL_OK;
}

int createContextCmd(Tcl_Interp *interp, struct SurfaceData *surfacePtr,
  int objc, Tcl_Obj *CONST objv[])
{
  char *name;
  name = ckalloc(strlen(Tcl_GetString(objv[0]))+4);
  strcpy(name, Tcl_GetString(objv[0]));
  strcat(name, "-cr");
  Tcl_CreateObjCommand(interp, name, contextObjCmd, surfacePtr, NULL);
  Tcl_SetObjResult(interp, Tcl_NewStringObj(name, -1));
  ckfree(name);
  return TCL_OK;
}

int surfaceObjCmd(ClientData data, Tcl_Interp *interp, int objc,
  Tcl_Obj *CONST objv[])
{
  struct SurfaceData *surfacePtr = (struct SurfaceData *)data;
  static const char *surfaceTypes[] = {
    "context", "toPng", "set_size",
    NULL
  };
  enum surfaceTypeIdx {
    CONTEXT_IX, TO_PNG_IX, SET_SIZE_IX
  };
  int index;
  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "command");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], surfaceTypes, "type", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  switch (index) {
    case CONTEXT_IX:
      return createContextCmd(interp, surfacePtr, objc, objv);
    case TO_PNG_IX:
      return toPng(interp, surfacePtr, objc, objv);
    case SET_SIZE_IX:
      return pdfSetSize(interp, surfacePtr, objc, objv);
  }
  return TCL_OK;
}

void surfaceObjCleanup(ClientData data) {
  struct SurfaceData *surfacePtr = (struct SurfaceData *)data;
  if (surfacePtr->context != NULL)
    cairo_destroy(surfacePtr->context);
  if (surfacePtr->surface != NULL)
    cairo_surface_destroy(surfacePtr->surface);
  surfacePtr->context = NULL;
  surfacePtr->surface = NULL;
}

int surfaceCreate(Tcl_Interp *interp, struct SurfaceState *statePtr,
    int objc, Tcl_Obj *CONST objv[])
{
  Tcl_HashEntry *entryPtr;
  struct SurfaceData *surfacePtr;
  char name[20];
  int new;
  static const char *surfaceTypes[] = {
    "argb32", "rgb24", "pdf", "svg",
    NULL
  };
  enum surfaceCmdIdx {
    ARGB32_IX, RGB24_IX, PDF_IX, SVG_IX};
  int index;
  double width;
  double height;
  char *filename;

  if (objc < 5) {
    Tcl_WrongNumArgs(interp, 2, objv, "surfaceType width height ?file?");
    return TCL_ERROR;
  }
  Tcl_GetDoubleFromObj(interp, objv[3], &width);
  Tcl_GetDoubleFromObj(interp, objv[4], &height);
  if (Tcl_GetIndexFromObj(interp, objv[2], surfaceTypes, "type", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  surfacePtr = (struct SurfaceData *)ckalloc(sizeof(struct SurfaceData));
  switch (index) {
    case ARGB32_IX:
      surfacePtr->surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32,
        width, height);
      break;
    case RGB24_IX:
      surfacePtr->surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,
        width, height);
      break;
    case PDF_IX:
      if (objc != 6) {
	Tcl_WrongNumArgs(interp, 5, objv, "file");
	return TCL_ERROR;
      }
      filename = Tcl_GetString(objv[5]);
      surfacePtr->surface = cairo_pdf_surface_create(filename, width, height);
      break;
    case SVG_IX:
      if (objc != 6) {
	Tcl_WrongNumArgs(interp, 5, objv, "file");
	return TCL_ERROR;
      }
      filename = Tcl_GetString(objv[5]);
      surfacePtr->surface = cairo_svg_surface_create(filename, width, height);
      break;
  }
  surfacePtr->context = cairo_create(surfacePtr->surface);
  statePtr->uid++;
  sprintf(name, "surface%d", statePtr->uid);
  entryPtr = Tcl_CreateHashEntry(&statePtr->hash, name, &new);
  Tcl_SetHashValue(entryPtr, (ClientData)surfacePtr);
  Tcl_CreateObjCommand(interp, name, surfaceObjCmd, (ClientData)surfacePtr,
    surfaceObjCleanup);
  Tcl_SetObjResult(interp, Tcl_NewStringObj(name, -1));
  return TCL_OK;
}

int surfaceDelete(Tcl_Interp *interp, struct SurfaceState *statePtr,
    int objc, Tcl_Obj *CONST objv[])
{
  Tcl_HashEntry *entryPtr;
  char *name;
  char *crname;

  name = Tcl_GetString(objv[2]);
  crname = ckalloc(strlen(name)+4);
  strcpy(crname, name);
  strcat(crname, "-cr");
  entryPtr = Tcl_FindHashEntry(&statePtr->hash, name);
  if (entryPtr == NULL)
    return TCL_OK;
  Tcl_DeleteCommand(interp, crname);
  Tcl_DeleteCommand(interp, name);
  Tcl_DeleteHashEntry(entryPtr);
  return TCL_OK;
}

int SurfaceCmd(ClientData data, Tcl_Interp *interp, int objc,
    Tcl_Obj *CONST objv[])
{
  struct SurfaceState *statePtr = (struct SurfaceState *)data;
  static const char* subCmds[] = {
    "create", "delete", NULL
  };
  enum surfaceCmdIdx {
    CREATE_IX, DELETE_IX
  };
  int index;

  if (objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
    return TCL_ERROR;
  }
  if (Tcl_GetIndexFromObj(interp, objv[1], subCmds, "option", 0, &index)
      != TCL_OK) {
    return TCL_ERROR;
  }
  switch (index) {
    case CREATE_IX:
      return surfaceCreate(interp, statePtr, objc, objv);
      break;
    case DELETE_IX:
      return surfaceDelete(interp, statePtr, objc, objv);
      break;
  }
  return TCL_OK;
}

void CairoCleanup(ClientData data) {
  struct SurfaceState *statePtr = (struct SurfaceState *)data;
  struct SurfaceData *surfacePtr;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;

  entryPtr = Tcl_FirstHashEntry(&statePtr->hash, &search);
  while (entryPtr != NULL) {
    Tcl_DeleteHashEntry(entryPtr);
    surfacePtr = (struct SurfaceData *)Tcl_GetHashValue(entryPtr);
    cairo_surface_destroy(surfacePtr->surface);
    entryPtr = Tcl_FirstHashEntry(&statePtr->hash, &search);
  }
}

int Cairotcl_Init(Tcl_Interp *interp) {
  struct SurfaceState *statePtr;

  if (Tcl_InitStubs(interp, "8.0", 0) == NULL) {
    return TCL_ERROR;
  }
  if (Tcl_PkgRequire(interp, "Tcl", "8.0", 0) == NULL) {
    return TCL_ERROR;
  }
  statePtr = (struct SurfaceState *)ckalloc(sizeof(struct SurfaceState));
  Tcl_InitHashTable(&statePtr->hash, TCL_STRING_KEYS);
  statePtr->uid = 0;
  Tcl_CreateObjCommand(interp, "cairo::surface", SurfaceCmd,
    (ClientData)statePtr, CairoCleanup);
  if (Tcl_PkgProvide(interp, "cairo", CAIROTCL_VERSION) != TCL_OK) {
    return TCL_ERROR;
  }
  Tcl_RegisterObjType(&PathType);
  return TCL_OK;
}

